from collections import deque

import numpy as np

# path = np.random.randint(0, 2, size=(5, 3))
# non_zero = np.where(path[:,1] != 0)[0]
# sel_rows = list(non_zero + 1) 
# ",".join([str(i) for i in sel_rows])

def conv_errs(test_errors, tol = 1e-4):

    opt_j = np.inf
    for j in range(1, len(test_errors)):
        if np.abs(test_errors[j] - test_errors[j-1])/test_errors[j] < tol:
            opt_j = j
            break

    return opt_j

def constraint_test_error(test_errors, path):
    start_j = 0
    for j in range(len(test_errors)):
        if j == 0:
            # first j is always zero. If not, it is
            # floating point error
            continue
        sel = get_modelselection(path, j)
        if sel:
            start_j = j
            break

    return test_errors[start_j:], start_j


def data_driven_lambda_k(test_errors, path, verbose = False, tol = 1e-4, ignore_conv = False):
    """
    Choose the best j based on the test errors

    it  constraints the test errors to the first non-zero selection
    imagine path has the following set of non-zero selections:
    idx: 0    1   2  |  3    4  ...
        [ 0,  0,  0, | x1,  x2, ...] -> path_nz
        [e1, e2, e3, | e4, e_m, ...] -> test_errors

    then the start_j = 3. The opt_j =  start_j + opt_j_c.

    if the opt_j_c is at e_m, then opt_j_c = 1 and opt_j = 3 + 1 = 4


    Parameters
    ----------
    test_errors : numpy.ndarray
        The test errors with shape (k,2)
        where the first column is the lambda values and the second column is the RMSE values
    path : numpy.ndarray
        The path of coefficients with shape (p,k) or (M,k) if it is from ISLE
        where p is the number of features and k is the number of lambda values
    verbose : bool, optional
        If True, the function will print the chosen j
        The default is False.
    tol : float, optional
        The tolerance for the convergence of the test errors
        The default is 1e-4.
    ignore_conv : bool, optional
        If True, the function will ignore the convergence of the test errors
        The default is False.
    Returns
    -------
    int
        The index of the best j
        This is the index of the lambda value that minimizes the test error
        or the index of the lambda value that converges to the test error
        depending on the convergence of the test errors.
        If the convergence is ignored, then it returns the index of the minimum test error.
    """
    # make sure the opt_j is not an empty selection
    test_errors_c, start_j = constraint_test_error(test_errors, path)

    min_err_j = np.argmin(test_errors_c) #O(k) = O(1) for fixed k
    conv_err_j = conv_errs(test_errors_c, tol=tol) #O(k) = O(1) for fixed k

    if conv_err_j < min_err_j and not ignore_conv:
        if verbose:
            print("lambda_k choosen by test error convergence", end= "")

        opt_j_c = conv_err_j
    else:
        if verbose:
            print("lambda_k choosen by minimum test error", end= "")
        opt_j_c = min_err_j

    return opt_j_c + start_j

def get_modelselection(path, j):
    """
    Get the number of non-zero coefficients for the j-th lambda value
    """
    if isinstance(path, np.ndarray):
        # beta_j is the j-th column of path
        # which is obtained from the j-th lambda value
        beta_j = path[:,j]
        # selection by elastic net
        return np.sum(beta_j != 0)
    
    else:
        # it is comming from the ISLE new_path
        # which is a list of splitting variables
        # indeces
        return len(path[j])

def choose_j(path, test_errors = None, factor = -1,
              verbose = False, tol = 1e-4, p = None,
              ignore_conv = False):
    """
    Choose the best j based on the path and test errors
    Parameters
    ----------
    path : numpy.ndarray
        The path of coefficients with shape (p,k) or (M,k) if it is from ISLE
        where p is the number of features and k is the number of lambda values
    test_errors : numpy.ndarray, optional
        The test errors with shape (k,2)
        where the first column is the lambda values and the second column is the RMSE values
    factor : float, optional
        The factor to choose the best j
        if factor is -1, then the function will return the index of the minimum test error
        if factor is between 0 and 1, then the function will return the index of the best j
        based on the number of non-zero coefficients
        The default is 1/2.
    verbose : bool, optional
        If True, the function will print the chosen j
        The default is False.
    tol : float, optional
        The tolerance for the convergence of the test errors
        The default is 1e-4.

    Returns
    -------
    int
        The index of the best j
    
    """

    # path has (p,k) shape, where p is the number of features
    # and k is the number of lambda values
    if factor == -1:
        # tests_errors contains two columns
        # the first one is the lambda values
        # the second one is the RMSE values
        # check 'calculate_test_errors' function
        
        # O(np) for obtain test_errors
        return data_driven_lambda_k(test_errors[:,1], path, verbose, tol, ignore_conv) # O(k) = O(1) for fixed k
        # return np.argmin(test_errors[:,1]) # O(k) = O(1) for fixed k
    
    else:
        if factor < 0 or factor > 1:
            raise ValueError('Factor must be between 0 and 1 if factor is not -1.')            
        
        if verbose:
            print(f"lambda_k choosen in function of factor ({factor}).")

        # recall p is the number of features
        k = path.shape[1] if isinstance(path, np.ndarray) else len(path)
        user_selection = np.round(p*factor).astype(int)            

        best_dist = np.inf
        best_j = 0
        for j in range(k):

            model_selection = get_modelselection(path, j)
            # distance between of desired number of non-zero
            # coefficients and the current number of non-zero
            tmp_dist = np.abs(model_selection - user_selection)
            if tmp_dist < best_dist:
                best_dist = tmp_dist
                best_j = j

        return best_j

def get_nonzero(path, j):

    if isinstance(path, np.ndarray):
        # stills returns an np.array
        return np.where(path[:,j] != 0)[0] # O(T^4)
    
    else:
        return path[j] # O(1)

def add_offset(beta_j_nz):
    """
    Add offset to the path
    and if beta_j_nz is an array
    then convert it to a list
    """
    if isinstance(beta_j_nz, np.ndarray):
        return list(beta_j_nz + 1)
    
    else:
        return [i + 1 for i in beta_j_nz]

def read_CT(CT_file):
    CT = np.loadtxt(CT_file, delimiter=',', skiprows=1, dtype=str)
    CT_spps = CT[:, :4]
    n_spps = len(np.unique(CT_spps))
    return CT_spps, n_spps

def row_selection(path, test_errors,  p = None,
                  factor = 1/2, inbetween = 0, 
                  check_spps = False, CT_file = None,
                  verbose = False, tol = 1e-4, 
                  ignore_conv = False):
    
    if check_spps:
        CT_spps, n_spps = read_CT(CT_file) # O(T^4)
        

    # path has (p,k) shape, where p is the number of features
    j_opt = choose_j(path, test_errors, factor = factor, verbose=verbose, tol=tol, p = p, ignore_conv = ignore_conv) # O(k) = O(1) for fixed k
    chosen_j = np.linspace(0, j_opt, 2 + inbetween,
                            endpoint=True, dtype=int) 

    taken = set()
    new_batches = deque()
    for j_int in chosen_j:

        # once it is integer,
        # it might be the case
        # that there are repeated j's
        if j_int in taken:
            continue
        
        # in any of the paths: elastic net or ISLE
        # the first column is all zeros
        if j_int == 0:
            taken.add(j_int)
            continue
    
        beta_j_nz = get_nonzero(path, j_int) # O(1) for isle

        if check_spps:
            # check on the number of species
            if len(np.unique(CT_spps[beta_j_nz,:])) < n_spps:
                taken.add(j_int)
                continue

        # plus one as Julia starts from 1
        # O(\rho T^4) for isle
        new_batches.append(  add_offset(beta_j_nz) )
        taken.add(j_int)
    
    return list(new_batches), j_opt

def write_rows(outfile, rows):

    with open(outfile, 'w') as f:
        for row in rows:
            # convert the row to a string
            f.write(",".join([str(b) for b in row]) + "\n")




import numpy as np
from numba import njit, float64, int64

@njit
def msoft_threshold( delta, lam, denom):
    
    if delta > lam:
        return (delta - lam) / denom
    
    elif delta < -lam:
        return (delta + lam) / denom
    
    else:
        return 0

# region epoch functions
@njit
def epoch_lasso(X, beta, lam, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps, s_new, diff):

    for j in chosen_ps:
        # j = 1
        b_old = beta[j]
        
        A2 = np.where(beta != 0)[0]
        # take out j from the array A2
        A2 = A2[A2 != j]
        tmp_X_T_X = X_T_X[j,:]

        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
        denom = c1 * Xj_T_Xj[j]
        beta[j] = msoft_threshold(delta, lam, denom)


        # beautiful rank 1 sums 
        if beta[j] != 0.0:
            s_new += X[:,j] * beta[j]

        diff_j = b_old - beta[j]
        if diff_j != 0.0:    
            diff += X[:,j] * diff_j


@njit
def epoch_lasso_v2(X: float64[:,:], beta: float64[:], lam: float64, r: float64[:], c1: float64,
                   n: int64, chosen_ps: int64[:], s_new: float64[:], s_diff: float64[:], zero_thresh: float64) -> None:

    denom = c1 * n
    # O(9*np) = O(np)
    for j in chosen_ps:
    
        b_old = beta[j]
        delta = c1 * ( np.dot(X[:,j], r) + n*b_old) # O(n)

        beta[j] = msoft_threshold(delta, lam, denom)
        # if beta[j] is too small, set it to 0
        if abs(beta[j]) < zero_thresh:
            beta[j] = 0.0


        diff_bj = b_old - beta[j]
        if diff_bj != 0.0:    
            Xj_diff_bj = X[:,j] * diff_bj # O(2n)

            r      += Xj_diff_bj # O(2n)
            s_diff += Xj_diff_bj # O(2n)

        # rank 1 sums 
        if beta[j] != 0.0:
            s_new += X[:,j] * beta[j] # O(2n)


@njit
def epoch_enet(X, beta, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps, lam_1_alpha, lam_alpha, s_new, diff):

    for j in chosen_ps:
        # j = 1
        b_old = beta[j]
        
        A2 = np.where(beta != 0)[0]
        # take out j from the array A2
        A2 = A2[A2 != j]
        tmp_X_T_X = X_T_X[j,:]

        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
        denom = (c1 * Xj_T_Xj[j]) + lam_1_alpha

        beta[j] = msoft_threshold(delta, lam_alpha, denom)


        # beautiful rank 1 sums 
        if beta[j] != 0.0:
            s_new += X[:,j] * beta[j]

        diff_j = b_old - beta[j]
        if diff_j != 0.0:    
            diff += X[:,j] * diff_j



@njit
def epoch_enet_v2(X: float64[:,:], beta: float64[:], lam_1_alpha: float64, lam_alpha: float64, r: float64[:], 
                  c1: float64, n: int64, chosen_ps: int64[:], s_new: float64[:], s_diff: float64[:], 
                  zero_thresh: float64) -> None:

    denom = (c1 * n) + lam_1_alpha
    for j in chosen_ps:

        b_old = beta[j]
        delta = c1 * ( np.dot(X[:,j], r) + n*b_old )

        beta[j] = msoft_threshold(delta, lam_alpha, denom)
        # if beta[j] is too small, set it to 0
        if abs(beta[j]) < zero_thresh:
            beta[j] = 0.0

        diff_bj = b_old - beta[j]
        if diff_bj != 0.0:    
            Xj_diff_bj = X[:,j] * diff_bj # O(2n)

            r      += Xj_diff_bj # O(2n)
            s_diff += Xj_diff_bj # O(2n)

        # rank 1 sums 
        if beta[j] != 0.0:
            s_new += X[:,j] * beta[j] # O(2n)

# endregion

# region update functions
@njit
def update_beta_lasso(beta, lam, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps):

    for j in chosen_ps:
        # j = 1
        # A2 = np.where(beta !=  0)[0]
        # # take out j from the array A2
        # A2 = A2[A2 != j]

        A2 = np.where(beta !=  beta[j])[0]

        tmp_X_T_X = X_T_X[j,:]

        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
        denom = c1 * Xj_T_Xj[j]
        beta[j] = msoft_threshold(delta, lam, denom)



@njit
def update_beta_lasso_v2(X: float64[:,:], beta: float64[:], lam: float64, r: float64[:], 
                         c1: float64, n: int64, chosen_ps: int64[:], zero_thresh: float64) -> None:

    denom = c1 * n
    # O(3*np) = O(np)
    for j in chosen_ps:

        b_old = beta[j]
        delta = c1 * ( np.dot(X[:,j], r) + n*b_old ) # O(n)

        beta[j] = msoft_threshold(delta, lam, denom)
        # if beta[j] is too small, set it to 0
        if abs(beta[j]) < zero_thresh:
            beta[j] = 0.0

        diff_bj = b_old - beta[j]
        if diff_bj != 0.0:
            r  += X[:,j] * diff_bj # O(3n)


@njit
def update_beta_enet(beta, c1, Xj_T_y, Xj_T_Xj, X_T_X, 
                     chosen_ps, lam_1_alpha, lam_alpha):
    for j in chosen_ps:
        # j = 1
        A2 = np.where(beta != 0)[0]
        # take out j from the array A2
        A2 = A2[A2 != j]

        tmp_X_T_X = X_T_X[j,:]

        delta = c1 * ( Xj_T_y[j] - np.dot(tmp_X_T_X[A2], beta[A2]) )
        denom = (c1 * Xj_T_Xj[j]) + lam_1_alpha

        beta[j] = msoft_threshold(delta, lam_alpha, denom)


@njit
def update_enet_v2(X: float64[:, :], beta: float64[:], lam_1_alpha: float64, lam_alpha: float64,
                   r: float64[:], c1: float64, n: int64, chosen_ps: int64[:], zero_thresh: float64) -> None:

    denom = (c1 * n) + lam_1_alpha
    for j in chosen_ps:

        b_old = beta[j]
        delta = c1 * ( np.dot(X[:,j], r) + n*b_old )

        beta[j] = msoft_threshold(delta, lam_alpha, denom)
        # if beta[j] is too small, set it to 0
        if abs(beta[j]) < zero_thresh:
            beta[j] = 0.0

        diff_bj = b_old - beta[j]
        if diff_bj != 0.0:
            r  += X[:,j] * diff_bj # O(3n)

# endregion

# region dual gap functions
def calculate_dual_gap(R, X, y, lam, beta):
    n = len(y)

    p_obj = (1/n) * np.linalg.norm(R)**2  + lam * np.linalg.norm(beta, ord=1)

    theta = (2/n) * R
    norm_inf = np.linalg.norm(X.T @ theta, ord=np.inf)
    theta_ = theta/max(lam, norm_inf)

    d_obj = (1/n)*( np.linalg.norm(y)**2 - np.linalg.norm(y - (lam*n/2)*theta_)**2 )

    w_d_gap = (p_obj - d_obj)*n/np.linalg.norm(y)**2

    return w_d_gap


def dualpa(X, y, R, lam, beta, ny2):

    n = len(y)
    c1 = 1/n

    rt = 2*c1*R
    
    # O(np)
    norm_inf = np.linalg.norm(X.T @ rt, ord=np.inf)
    c2 = lam/max(norm_inf, lam)
    
    # O(n + p)
    d_gap = (1/n)*np.dot(R, (1 + c2**2)*R - 2*c2*y) + lam*np.linalg.norm(beta, ord=1)

    return d_gap*n/ny2


def dualpa_elnet(R, X, y, beta, lam_1_alpha, lam_alpha, ny2):
    
    if lam_1_alpha == 0:
        return dualpa(X, y, R, lam_alpha, beta, ny2)

    n = len(y)
    b_norm_sq = np.dot(beta, beta)

    rt = (2/n) * R
    norm_inf = np.linalg.norm(X.T @ rt  - lam_1_alpha*beta, ord=np.inf)

    c2 = lam_alpha/max(norm_inf, lam_alpha)
    c3 = 1 + c2**2

    d_gap = (1/n)*np.dot(R, c3*R - 2*c2*y) +\
            lam_alpha*np.linalg.norm(beta, ord=1) +\
            c3*(lam_1_alpha/2)*b_norm_sq

    return d_gap*n/ny2

# endregion


from matplotlib import pyplot as plt
import numpy as np
from collections import deque
from copy import deepcopy
from sklearn import tree
from sklearn.tree import DecisionTreeRegressor
import sklearn.model_selection


# generate 100 random numbers

X = np.random.rand(442, 1000)
y = np.random.rand(442) 

# random state is shared across all trees
rng = np.random.RandomState(12038)
model = DecisionTreeRegressor(max_leaf_nodes=6, max_depth=5, max_features=0.5)
estimators = deque()
for i in range(100):
    # this generates a random based in function
    # of the random state rng
    model.set_params(random_state=rng)
    model.fit(X, y)
    estimators.append(deepcopy(model)) #O(d)
    # indx = set(model.tree_.feature[model.tree_.feature != -2])
    # print(indx)

estimators = np.array(list(estimators))
k = 200
path = rng.randint(0, 2, size=(len(estimators), k))


m = estimators[0]
tree.plot_tree(m, proportion=True)
plt.show()

m.tree_.feature[m.tree_.feature != -2]


sklearn.model_selection.GridSearchCV

# path = get_new_path(estimators, path, p)
# picked_batches = select_path(path, CT_spps, test_errors, n_spps, args.factor, args.inbetween)
# write_batches(picked_file, picked_batches)
# ---------------------^^  In progress ^^ --------------------------------

# # disjoint batches creation
# batches = make_batches(path, CT_spps, n_spps)
# if args.window <= 1:
#     write_batches(batch_file, batches)
# else:
#     new_batches = agglomerate_batches(batches, window=args.window)
#     write_batches(batch_file, new_batches)
# if not args.nwerror and args.verbose:
#     # print(test_errors)
#     j_min = np.argmin(test_errors[:,1])
#     new_path_j = path[:,j_min]
#     min_err_sel = np.sum(new_path_j != 0)
#     print("Number of rows selected at min error: ", min_err_sel)

# --------------------------------------------------------------------------

# only in the case of ISLE
def get_new_path(estimators, path):
    """
    Takes the splitting variables from the trees
    """

    estimators = np.array(estimators) # O(M)
    new_path = deque()
    for j in range(path.shape[1]):
        if j == 0:
            continue

        coeffs = path[:,j]
        coeffs_logic = coeffs != 0

        # filter for active trees
        tmp_ensemble = estimators[coeffs_logic] # O(M)

        I_k = set()
        # O(M)
        for m in tmp_ensemble: 
            I_k_m = set(m.tree_.feature[m.tree_.feature != -2])
            I_k |= I_k_m
            
        new_path.append(I_k)

    return new_path



def choose_j(path, test_errors = None, factor = 1/2):
    """
    Choose the best j based on the path and test errors
    Parameters
    ----------
    path : numpy.ndarray
        The path of coefficients with shape (p,k)
        where p is the number of features and k is the number of lambda values
    test_errors : numpy.ndarray, optional
        The test errors with shape (k,2)
        where the first column is the lambda values and the second column is the RMSE values
    factor : float, optional
        The factor to choose the best j
        if factor is -1, then the function will return the index of the minimum test error
        if factor is between 0 and 1, then the function will return the index of the best j
        based on the number of non-zero coefficients
        The default is 1/2.

    Returns
    -------
    int
        The index of the best j
    
    """

    # path has (p,k) shape, where p is the number of features
    # and k is the number of lambda values
    if factor == -1 and test_errors is not None:
        # tests_errors contains two columns
        # the first one is the lambda values
        # the second one is the RMSE values
        # check 'calculate_test_errors' function
        
        # O(np) for obtain test_errors
        return np.argmin(test_errors[:,1]) # O(k) = O(1) for fixed k
    
    else:
        if factor < 0 or factor > 1:
            raise ValueError('Factor must be between 0 and 1 if nwerror is false and factor is not -1.')

        # recall p is the number of features
        p,k = path.shape
        user_selection = np.round(p*factor).astype(int)

        best_dist = np.inf
        best_j = 0
        for j in range(k):
            # beta_j is the j-th column of path
            # which is obtained from the j-th lambda value
            beta_j = path[:,j]
            # selection by elastic net
            model_selection = np.sum(beta_j != 0)
            # distance between of desired number of non-zero
            # coefficients and the current number of non-zero
            tmp_dist = np.abs(model_selection - user_selection)
            if tmp_dist < best_dist:
                best_dist = tmp_dist
                best_j = j

        return best_j


def select_path(path, CT_spps, test_errors, n_spps = 15, 
                factor = 1/2, inbetween = 0, 
                isle = False):
    
    """
    Select the path based on the test errors and the number of species
    Parameters
    ----------
    path : numpy.ndarray
        The path of coefficients with shape (p,k)
        where p is the number of features and k is the number of lambda values
        or (M,k) where M is the number of trees
    CT_spps : numpy.ndarray
        The concordance table with shape (n_spps, p)
        where n_spps is the number of species and p is the number of features
    test_errors : numpy.ndarray
        The test errors with shape (k,2)
        where the first column is the lambda values and the second column is the RMSE values
    n_spps : int, optional
        The number of species to be selected
        The default is 15.
    factor : float, optional
        The factor to choose the best j
        if factor is -1, then the function will return the index of the minimum test error
        if factor is between 0 and 1, then the function will return the index of the best j
        based on the number of non-zero coefficients
        The default is 1/2.
    inbetween : int, optional
        The number of inbetween j's to be selected
        The default is 0.
    isle : bool, optional
        If True, then the path is a M x k matrix
        where M is the number of trees and k is the number of lambda values
        The default is False.

    Returns
    -------
    deque
        The indices of the selected j's
        withn offset of 1
    """

    j_opt = choose_j(path, test_errors, factor = factor)
    chosen_j = np.linspace(0, j_opt, 2 + inbetween, 
                           endpoint=True, dtype = int)

    taken = set()
    new_batches = deque()
    for j in chosen_j:
        # it might be the case
        # that there are repeated j's
        if j in taken:
            continue

        if j == 0:
            taken.add(j)
            continue
        
        if not isle:
            # column vectors directly 
            # represent a feature or
            # row in the concordance table
            # i.e., path is a p x k matrix
            I = np.where(path[:,j] != 0)[0]
        else:
            # otherwise, the column represent the weights
            # for regression trees. Each tree used features.
            # TODO: add indeces from tree's splitting variables
            # in this case, path is M x k matrix

            # I has to be a list or an array of indices

            pass

        # check on the number of species
        # O(\rho T^4)
        if len(np.unique(CT_spps[I,:])) < n_spps:
            taken.add(j)
            continue

        # indices in python start from 0 but 
        # julia starts from 1. Julia is
        # used to read the data
        new_batches.append(I + 1) 
        taken.add(j)
    
    return new_batches


new_batches = select_path(path, path, None, n_spps = 15, 
                factor = 1/2, inbetween = 0, 
                isle = False)

min_err_sel = len(new_batches[-1])
print("Number of rows selected at min error: ", min_err_sel)





cv_results_report_file = "/Users/ulises/Desktop/ABL/software/experiments_qsin/empirical_hyper_leaf2.txt"

cv_results_report_l2 = {}
cv_results_report_l3 = {}

with open(cv_results_report_file, 'r') as f:
    for line in f.readlines():
        line = line.strip().split(",")
        
        alpha = float(line[1].replace("alpha: ", ""))
        nj = float(line[2].replace("eta: ", ""))
        vj = float(line[3].replace("nu: ", ""))
        lj = int(line[4].replace("leaves: ", ""))

        err = float(line[5].replace("err: ", ""))
        theta = (alpha, nj, vj)

        if lj == 2:
            if theta not in cv_results_report_l2:
                cv_results_report_l2[theta] = err/5
            else:
                cv_results_report_l2[theta] += err/5

        else:  
            if theta not in cv_results_report_l3:
                cv_results_report_l3[theta] = err/10
            else:
                cv_results_report_l3[theta] += err/10
        


sorted_l2 = sorted(cv_results_report_l2.items(), key=lambda x: x[1])
sorted_l3 = sorted(cv_results_report_l3.items(), key=lambda x: x[1])



a = [0/2,
4/2,
6/2,
8/2,
10/2,
12/2,
14/2,
16/2,
18/2,
20/2,
22/2,
24/2,
26/2,
28/2,
30/2,
32/2,
34/2,
36/2,
38/2,
40/2,
42/2,
44/2,
46/2,
48/2,
50/2,
56/2,
58/2,
60/2,
62/2,
64/2,
66/2,
68/2,
70/2,
72/2,
74/2,
76/2,
78/2,
80/2,
82/2,
84/2,
86/2,
88/2]

ran = set([ int(i) for i in a])
sub = set([ int(i) for i in range(0,51)])


for i in sub - ran:
    print(f"sbatch nl_submitter_{i}.sh")






import sys
import time
import random

import numpy as np

def max_lambda(X, y, alpha = None):
    # 1/n * max( |X^T * y| )
    # SLS book page 114
    n,_ = X.shape

    if alpha is not None:
        c1 = 2 / (alpha * n)
    else:
        c1 = 2 / n

    return c1 * np.max( np.abs( np.matmul(X.T, y) ) )


def get_lambdas(alpha, X_train, y_train, K, epsilon, verbose=False):

    # max_lambda = O(nT^4) operations, under the budget of
    # simulations, or ISLE ensemble, it is negligible
    max_lam = max_lambda(X_train, y_train, alpha=alpha)
    min_lam = max_lam * epsilon

    if verbose:
        print("Lambda range: ", min_lam, max_lam)

    lambdas =  np.logspace(np.log10(min_lam), np.log10(max_lam), K, endpoint=True)[::-1]
    return lambdas

def get_alpha_max_lam(X_train, y_train, alphas):
    
    all_max_lams = -np.inf
    all_max_lams_alpha_indx = -1
    for i,a in enumerate(alphas):
        tmp_a = max_lambda(X_train, y_train, alpha=a)
        if tmp_a > all_max_lams:
            all_max_lams = tmp_a
            all_max_lams_alpha_indx = i
            
    return alphas[all_max_lams_alpha_indx]


def rmse(y, X, B):
    return np.sqrt( np.mean( ( y - X.dot(B) )**2 ) )

def split_data(X,y,num_test, seed = 123):

    random.seed(seed)
    n,_ = X.shape

    test_idx  = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    return X_train, X_test, y_train, y_test

def write_test_errors(prefix, errors, lambdas):
    """
    Calculate the test errors for each lambda value in the path

    Parameters
    ----------
    prefix : str
        The prefix for the output file

    errors : list
        The list of RMSE values for each lambda

    lambdas : list
        The list of lambda values

    Returns
    -------
    numpy.ndarray
        The test errors, which is a 2D array with the first column
        indicating the lambda value and the second column indicating
        the RMSE value

    """
    test_errors = np.zeros((len(errors), 2))
    for j, rmse_j in enumerate(errors):
        test_errors[j, :] = [lambdas[j], rmse_j]

    np.savetxt(prefix + "_testErrors.csv",
                test_errors,
                delimiter=',',
                comments='')

    return test_errors

def progressbar(it, prefix="", size=60, out=sys.stdout): 
    """
    ref: https://stackoverflow.com/a/34482761
    Progress bar for loops

    """ 
    count = len(it)
    start = time.time() # time estimate start
    def show(j):
        x = int(size*j/count)
        # time estimate calculation and string
        remaining = ((time.time() - start) / j) * (count - j)        
        mins, sec = divmod(remaining, 60) # limited to minutes
        time_str = f"{int(mins):02}:{sec:03.1f}"
        print(f"{prefix}[{u'█'*x}{('.'*(size-x))}] {j}/{count} Est wait {time_str}", end='\r', file=out, flush=True)
    show(0.1) # avoid div/0 
    for i, item in enumerate(it):
        yield item
        show(i+1)
    print("\n", flush=True, file=out)


def standardize_Xy(X_train, y_train, X_test = None, y_test = None, nstdy = False):
    """
    Standardize the data
    """
    x_u = np.mean(X_train, axis = 0)
    x_sd = np.std(X_train, axis = 0)
    
    zero_sd_indices = x_sd == 0 # O(p)
    if np.any(zero_sd_indices): # O(p)
        x_sd[zero_sd_indices] = 1 # O(p)

    y_u = np.mean(y_train)
    y_sd = np.std(y_train)

    if y_sd == 0:
        y_sd = 1

    X_train = (X_train - x_u) / x_sd
    y_train = y_train - y_u if nstdy else (y_train - y_u) / y_sd

    if X_test is not None:
        X_test = (X_test - x_u) / x_sd
        y_test = y_test - y_u if nstdy else (y_test - y_u) / y_sd

    return X_train, X_test, y_train, y_test

def standardize_X(X_train,  X_test = None):
    """
    Standardize the data
    """
    x_u = np.mean(X_train, axis = 0)
    x_sd = np.std(X_train, axis = 0)
    
    x_sd_zero = x_sd == 0
    if np.any(x_sd_zero):
        x_sd[x_sd_zero] = 1

    X_train = (X_train - x_u) / x_sd

    if X_test is not None:
        X_test = (X_test - x_u) / x_sd

    return X_train, X_test




import random

import numpy as np
from qsin.utils import progressbar
from qsin.CD_dual_gap import (epoch_lasso, update_beta_lasso, dualpa, 
                              epoch_enet , update_beta_enet , dualpa_elnet)

def sparse_XB(X, B):
    return np.matmul(X, B)

class Lasso:
    """
    lasso model.
    It assumes that the input data has been standardized
    """
    def __init__(self, 
                 max_iter=300, 
                 lam=0.1,
                 prev_lam = None,
                 fit_intercept  = True,
                 warm_start = False,
                #  beta = None,
                 tol = 0.001,
                 init_iter = 1,
                 copyX = False,
                 seed = 123,
                 **kwargs):
        
        self.max_iter = max_iter
        self.lam = lam
        self.prev_lam = prev_lam
        self.tol = tol
        self.warm_start = warm_start
        self.seed = seed

        self.fit_intercept = fit_intercept
        self.beta = np.array([])


        self.intercept = 0
        
        self.init_iter = init_iter
        self.copyX = copyX

        self.X = np.array([])
        self.X_T_X = np.array([])
        self.Xj_T_Xj = np.array([])
        self.Xj_T_y = np.array([])
        self.ny2 = np.array([])

        self.y = np.array([])

        self.x_offset = np.array([])
        self.y_offset = 0.0

        self.x_scale = np.array([])
        self.y_scale = 1

        
    def initial_iterations(self, c1, Xj_T_y, Xj_T_Xj, X_T_X, all_p):

        for _ in range(self.init_iter):
            self.update_beta(c1, Xj_T_y, Xj_T_Xj, X_T_X, all_p)

    def initial_active_set(self, X, y, c1, Xj_T_y, Xj_T_Xj, X_T_X, all_p):

        if self.prev_lam is None:
            # few iterations of coordinate descent
            self.initial_iterations(c1, Xj_T_y, Xj_T_Xj, X_T_X, all_p)
            # we define an active set A as the set of indices
            A = np.where(self.beta != 0)[0]

        else:
            A = self.sequential_strong_rule(X, y, c1, all_p)

        return A
    
    def sparse_XB(self, X):
        """
        This function returns the sparse matrix
        X[:, A] * B[A], where A is the index set of
        non-zero coefficients
        """
        # A = np.where(self.beta != 0)[0]
        # return np.matmul(X[:, A], self.beta[A])

        return np.matmul(X, self.beta)
        # return sparse_XB(X, self.beta)
    
    def set_Xy(self, X,y):
        """
        set X

        This is used under the 
        following logic

        if copy and x == 0:
            set_X
        elif not copy and x == 0:
            set_X (overwrite)
        elif copy and x != 0:
            do nothing
        elif not copy and x != 0:
            set_X (overwrite)
        """

        if self.copyX and len(self.X) != 0:
            pass

        else:
            self.X = X.copy()
            self.y = y.copy()
            self.center_Xy()

            # optimal for dot products in CD
            self.X_T_X = np.asfortranarray( np.matmul(self.X.T, self.X) )
            self.Xj_T_Xj = np.diag(self.X_T_X)
            self.Xj_T_y = np.matmul(self.X.T, self.y)
            self.ny2 = np.dot(self.y, self.y)

    def center_Xy(self):
        """
        center X and y
        """
        self.x_offset = np.mean(self.X, axis=0, dtype=self.X.dtype) # O(np)
        self.x_scale  = np.std(self.X , axis=0, dtype=self.X.dtype) # O(np)
        
        zero_sd_indices = self.x_scale == 0 # O(p)
        if np.any(zero_sd_indices): # O(p)
            self.x_scale[zero_sd_indices] = 1 # O(p)

        self.X = (self.X - self.x_offset) / self.x_scale # O(np)

        self.y_offset = np.mean(self.y) # O(n)
        self.y -= self.y_offset # O(n)


    def dual_gap(self, y):
        return dualpa(self.X, y, self.lam, self.beta, self.ny2)
    
    def fit(self, X, y):
        # X = X_train
        # y = y_train
        # self.max_iter = 10000
        # self.lam = 10
        # self.set_params(max_iter = 1000, lam = 300)
        
        n, p = X.shape
        c1 = 2 / n

        self.set_Xy(X,y)

        
        if not self.warm_start:
            # He-styled 
            # initialization 
            # of the coefficients
            # np.random.seed(self.seed)
            # self.beta = np.random.normal(0, np.sqrt(2/p), size=p) 

            self.beta = np.zeros(p, dtype=self.X.dtype, order='F')  
         
        # few iterations of coordinate descent
        all_p = np.array(range(p))
        A = self.initial_active_set(self.X, y, c1, 
                                    self.Xj_T_y, 
                                    self.Xj_T_Xj, 
                                    self.X_T_X, 
                                    all_p)
        

        left_iter = self.max_iter - self.init_iter
        for i in range(left_iter):
            # i = 1
            diff = np.zeros(n)
            
            # s_old = np.matmul( self.X[:, A], self.beta[A] )
            s_new = np.zeros(n)
            self.cd_epoch(c1, A, s_new, diff)

            # self.update_beta(c1, self.Xj_T_y, self.Xj_T_Xj, self.X_T_X, A)
            # s_new = np.matmul( self.X[:, A], self.beta[A] )
            # s_new =  self.beta

            # diff = s_new - s_old
            max_updt = np.max(np.abs(diff))
            w_max = np.max(np.abs(s_new))

            # max_updt = np.linalg.norm(s_new - s_old, ord=np.inf)
            # w_max = np.linalg.norm(s_new, ord=np.inf)
            

            if w_max == 0 or max_updt/w_max < self.tol:
            # if np.linalg.norm(s_new - s_old) < self.tol:
                # print("Max updates: ", max_updt)

                # A_ = set(all_p) - set(A)
                A_ = np.setdiff1d(all_p, A)

                #TODO: streamline updt. beta and exclusion test
                self.update_beta(c1, 
                                 self.Xj_T_y, 
                                 self.Xj_T_Xj,
                                 self.X_T_X, A_)
                A_new = self.exclusion_test(self.X, y, c1, A, all_p)
                

                if len(A_new) == 0:
                    # it means that all
                    # coefficients follow the
                    # KKT conditions
                    # print('kkt, finished at iteration: ', i)
                    break

                else:

                    # R = y - self.sparse_XB(self.X)
                    # w_d_gap = dualpaR(R, self.X, y, self.lam, self.beta)
                    w_d_gap = self.dual_gap(self.y)
                    # w_d_gap = dualpa(self.X, y, self.lam, self.beta)

                    if w_d_gap < self.tol:
                        # print('dual, finished at iteration: ', i)
                        break

                    else:
                        A = np.concatenate((A, A_new))
                        A.sort(kind = 'mergesort')

        if i == left_iter - 1:
            # if the iterations reach this point,
            # it means that there is still an active set.
            # then, the model did not converge
            print("Model did not converge")

        if self.fit_intercept:
            self.beta /= self.x_scale
            self.intercept = self.y_offset - np.dot(self.x_offset, self.beta) # O(p)
            # print("Intercept: ", self.intercept)

    def cd_epoch(self, c1, chosen_ps, s_new, diff):
        epoch_lasso(self.X, self.beta, self.lam, c1,
                     self.Xj_T_y, self.Xj_T_Xj, self.X_T_X, 
                     chosen_ps, s_new, diff)

    def update_beta(self, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps):
        update_beta_lasso(self.beta, self.lam, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps)

        # for j in chosen_ps:
        #     # j = 1
        #     A2 = np.where(self.beta != 0)[0]
        #     # take out j from the array A2
        #     A2 = A2[A2 != j]

        #     delta = c1 * ( Xj_T_y[j] - np.dot(X_T_X[j, A2], self.beta[A2]) )
        #     denom = c1 * Xj_T_Xj[j]
        #     self.beta[j] = self.soft_threshold(delta, self.lam, denom)

    def exclusion_test(self, X, y, c1, A, all_p):
        """
        Exclusion test (SLS book, page 114)

        The optimization problem dictates that:

        X^T * r = lam * gamma
        
        gamma being partial of ||B||_1 and whose
        subgradient is given by:

        gamma_j = + 1     if beta_j > 0 (sign of beta_j)\\
        gamma_j = - 1     if beta_j < 0 (sign of beta_j)\\
        gamma_j = (-1, 1) if beta_j = 0 (sign of 0, undefined)\\

        Since the range of gamma_j is between -1 and 1,\\
        then the range of lam * gamma_j is between -lam and lam.\\
        So, the range of Xj^T * r is between -lam and lam.

        Leading the following set of inequalities:
        -lam <= Xj^T * r <= lam

        which from:\\
        lam >= Xj^T * r\\
        lam >= -Xj^T * r

        Implies: 
        lam >= |Xj^T * r|

        Now, we can define the exclusion test as:
        lam > |Xj^T * r|
        whose strict inequiality test for beta_j = 0
        as +1 and -1 are not in the range of the subgradient

        If we pass this test over the omited  variables
        that were supposed to be zero and fails, it means
        they were actually non-zero and should be included
        in the active set.
        """
        # A_ = set(all_p) - set(A)
        A_ = np.setdiff1d(all_p, A)

        if len(A_) == 0:
            return A_

        r = y - self.sparse_XB(X)

        # A_ = np.array(list(A_))
        # exclusion test (SLS book, page 114) based on
        # KTT conditions
        e_test = c1 * np.abs( np.matmul( X.T[A_, :], r) ) >= self.lam
        # those that are in the new active set
        # are those who did pass the KTT conditions
        return A_[e_test]

    def sequential_strong_rule(self, X, y, c1, all_p):
        """
        rule based on the dual polytope projection 
        and futher modified by the strong rule
        showed at SLS book page 128
        """
        A_ = np.array(all_p)

        r = y - self.sparse_XB(X)

        # sequential strong rule (SLS book, page 128)
        e_test = c1 * np.abs( np.matmul( X.T[A_, :], r) ) >= 2*self.lam - self.prev_lam
        return A_[e_test]
    
    def predict(self, X):
        if self.beta is None:
            raise Exception("Model has not been fitted yet.")
        
        return self.sparse_XB(X) + self.intercept
    
    def score(self, X, y):
        y_pred = self.predict(X)
        return np.sqrt( np.mean( (y - y_pred)**2 ) )
    
    def set_params(self, **params):
        if "max_iter" in params:
            self.max_iter = int(params["max_iter"])

        if "lam" in params:
            self.lam = params["lam"]

        if "intercept" in params:
            self.intercept = params["intercept"]

        if "tol" in params:
            self.tol = params["tol"]

        if "warm_start" in params:
            self.warm_start = params["warm_start"]

        if "beta" in params:
            self.beta = params["beta"]

        if "prev_lam" in params:
            self.prev_lam = params["prev_lam"]

    def get_params(self):
        return {'max_iter': self.max_iter, 
                'lam': self.lam, 
                'intercept': self.intercept, 
                'tol': self.tol}
    
    def set_beta(self, beta):
        self.beta = beta

class ElasticNet(Lasso):
    def __init__(self, 
                 max_iter=300, 
                 alpha = 0.5, 
                 lam=0.1, 
                 prev_lam=None, 
                 fit_intercept=True, 
                 warm_start=False,
                #  beta=None,
                 tol=0.001, 
                 **kwargs):
        super().__init__(max_iter, lam, prev_lam, fit_intercept, warm_start,  tol, **kwargs)

        self.alpha = alpha
        self.lam = lam

        # when lam or alpha
        # change, these values are
        # updated at set_params
        self.lam_alpha = alpha * lam
        self.lam_1_alpha = (1 - alpha) * lam

    def set_lam_alpha(self, lam, alpha):
        self.lam_alpha = alpha * lam
        self.lam_1_alpha = (1 - alpha) * lam

    def set_params(self, **params):
        if "max_iter" in params:
            self.max_iter = int(params["max_iter"])

        if "lam" in params:
            self.lam = params["lam"]

            if "alpha" in params:
                self.alpha = params["alpha"]
            
            self.set_lam_alpha(self.lam, self.alpha)
            # self.lam_alpha = self.alpha * self.lam
            # self.lam_1_alpha = (1 - self.alpha) * self.lam

        if 'alpha' in params:
            self.alpha = params['alpha']

            if "lam" in params:
                self.lam = params["lam"]

            self.set_lam_alpha(self.lam, self.alpha)
            # self.lam_alpha = self.alpha * self.lam
            # self.lam_1_alpha = (1 - self.alpha) * self.lam

        if "intercept" in params:
            self.intercept = params["intercept"]

        if "tol" in params:
            self.tol = params["tol"]

        if "warm_start" in params:
            self.warm_start = params["warm_start"]

        if "beta" in params:
            self.beta = params["beta"]

        if "prev_lam" in params:
            self.prev_lam = params["prev_lam"]

    def cd_epoch(self, c1, chosen_ps, s_new, diff):
        epoch_enet(self.X, self.beta, c1,
                   self.Xj_T_y, self.Xj_T_Xj, self.X_T_X, 
                   chosen_ps, self.lam_1_alpha, self.lam_alpha, s_new, diff)


    def update_beta(self, c1, Xj_T_y, Xj_T_Xj, X_T_X, chosen_ps):
        update_beta_enet(self.beta, c1, Xj_T_y, Xj_T_Xj,
                                     X_T_X, chosen_ps, self.lam_1_alpha,
                                     self.lam_alpha)

    def dual_gap(self, y):
        
        R = y - self.sparse_XB(self.X)
        return dualpa_elnet(R, self.X, y, self.beta, self.lam_1_alpha, self.lam_alpha, self.ny2)
        # return duality_gap_elnet(R, self.X, y, 
        #                           self.beta, self.lam_1_alpha,
        #                             self.lam_alpha)
    
    def exclusion_test(self, X, y, c1, A, all_p):
        # A_ = set(all_p) - set(A)
        A_ = np.setdiff1d(all_p, A)

        if len(A_) == 0:
            return A_

        r = y - self.sparse_XB(X)

        # A_ = np.array(list(A_))
        # exclusion test (SLS book, page 114) based on
        # KTT conditions

        # Xj_T_r = np.matmul( X.T[A_, :], r)

        if self.lam_1_alpha == 0:
            elastic_term = 0
        else:
            elastic_term = (1/c1) * self.lam_1_alpha * self.beta[A_]

        elastic_test = c1 * np.abs(np.matmul( X.T[A_, :], r) - elastic_term) >= self.lam_alpha
        # those that are in the new active set
        # are those who did pass the KTT conditions
        return A_[elastic_test]

def k_fold_cv(X, y, model, num_folds):
    n, p = X.shape
    fold_size = n // num_folds
    mse_sum = 0

    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]

        model.fit(X_train, y_train)
        # y_pred = lasso.predict(X_test)
        mse_sum += model.score( X_test, y_test )

    return mse_sum / num_folds

def k_fold_cv_random(X, y, 
                     model, 
                     params,
                     num_folds = 3, 
                     sample = 100,
                     verbose = False,
                     seed = 123,
                     warm_starts = False
                     ):
    
    # model = Lasso()
    # X = X_train
    # y = y_train
    
    np.random.seed(seed=seed)
    
    all_params = params.keys()
    tested_params = np.ones((sample, len(params)))

    for n,k in enumerate(all_params):
        tested_params[:,n] = np.random.choice(params[k], sample)
    
    if warm_starts:
        # check index where 'lam' is in 
        # all_params
        idx = list(all_params).index('lam')
        # sort the tested_params by using the idx in the decreasing
        # order
        tested_params = tested_params[tested_params[:,idx].argsort()[::-1]]
        # tested_params = tested_params[tested_params[:,idx].argsort()]
    
    all_errors = []
    for vec in tested_params:
        # vec = tested_params[1]
        tmp_params = dict(zip(all_params, vec))

        if warm_starts and len(model.beta):
            model.set_params(**tmp_params, 
                             warm_start=True, 
                             beta=model.beta)
        else:
            model.set_params(**tmp_params)


        tmp_err = k_fold_cv(X, y, model, num_folds)
        all_errors.append([tmp_params, tmp_err])

        if verbose:
            print('Error: %s, tested params: %s' % (tmp_err, vec))

    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]
    if verbose:
        print("CV score: ", best_[1])

    return best_[0]


def lasso_path(X_train, y_train, params, model, print_progress = True):
    """
    compute the lasso path based on the training set
    and  with errors based on the test set
    """
    # model = Lasso()
    # X = X_train
    # y = y_train
    # params = {'lam': np.logspace(-2, max_lambda(X,y, alpha), 3)}

    # if X_test is None and y_test is None:
    #     X_test = X_train
    #     y_test = y_train


    _,p = X_train.shape
    lams = params['lam']

    # errors = np.zeros(len(lams))
    path = np.ones((p, len(params['lam'])))

    model.set_params(lam=lams[0])
    model.fit(X_train, y_train)

    path[:,0] = model.beta

    if print_progress:
        index_set = progressbar(range(1, len(lams)), "Computing lasso path: ", 40)
        
    else:
        index_set = range(1, len(lams))
    
    for i in index_set:
    # for i in range(1, len(lams)):

        model.set_params(lam = lams[i],
                          warm_start = True, 
                          prev_lam = None
                          )
            
        model.fit(X_train, y_train)
        path[:,i] = model.beta

    return path


def theta(path_i):

    path = np.copy(path_i)
    theta_path = np.zeros(path.shape)

    for i in range(path.shape[1]):
        tmp_path = path[:,i]
        logic_gate = tmp_path != 0

        tmp_path[logic_gate] = 1
        theta_path[:,i] = tmp_path

    return theta_path

def split_data(X,y,num_test, seed = 123):

    random.seed(seed)
    n,_ = X.shape

    test_idx  = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    return X_train, X_test, y_train, y_test

def get_ZO(path, spps_array, n_spps):

    Z = np.zeros((n_spps, path.shape[0]))
    for i in range(path.shape[0]):
        Z[:,i][spps_array[i,:] - 1] = 1

    theta_path = theta(path)
    return Z @ theta_path

def get_non_zero_coeffs(path, ZO, thresh = 0.5):
    n_features = path.shape[0]
    string_version = []
    non_zero_coeffs = []
    for j in range(path.shape[1]):
        # j = 10
        path_j = path[:,j]
        ZO_j = ZO[:,j]

        # checking if all species are
        # covered by the selected coefficients
        if np.any(ZO_j == 0):
            continue

        non_zero = path_j != 0

        if np.all(non_zero):
            break

        if np.sum(non_zero) <= thresh*n_features:
            non_zero_idx = list(np.where(non_zero)[0])
            non_zero_idx_str = str(set(np.sort(non_zero_idx)))

            if non_zero_idx_str not in string_version:
                string_version.append(non_zero_idx_str)
                non_zero_coeffs.append(non_zero_idx)

    return non_zero_coeffs



import numpy as np
import multiprocessing as mp
from collections import deque

def error_fn(theta_j, base_model = None, i = None, 
             X_train_t = None, X_test_t = None, 
             y_train_t = None, y_test_t = None,
             verbose = True, seed = None):
    """
    Compute the error for a given set of hyperparameters.
    """
    rng = np.random.RandomState(seed)
    nj, vj, lj, alpha = theta_j
    
    # scaled n_j such that the sub sample size
    # [nj_p * N_{I\I_k} ] = [ nj * n ] 
    # matches the desired number of rows that n_j induces 
    # into the complete training set.
    N_train = X_train_t.shape[0]

    # if nj is not None:
    #     nj_p = nj*n/N_train

    base_model.set_params(
        eta = nj,
        nu = vj,
        max_leaves = lj,
        alpha = alpha,
        rng = rng,
        verbose = False,
    )
    base_model.fit(X_train_t, y_train_t)
    # sum of squared errors:
    tmp_errors = base_model.score(X_test_t, y_test_t, metric = 'sse')

    if verbose:
        min_err = np.round(tmp_errors[-1], 4)
        fold_message  = f"Fold: {i}, alpha: {alpha}"
        fold_message += f", eta: {nj} (N for trees: {int(nj*N_train)})" if nj else ""
        fold_message += f", nu: {vj}" if vj else ""
        fold_message += f", leaves: {lj}" if lj else ""
        fold_message += f", Err: {min_err}"
        print(fold_message)

    return (tmp_errors, theta_j)


def get_parallel_errors(base_model, X_train, y_train, full_grid, 
                        rng, num_folds, ncores,
                        verbose = False, tmp_seed = None):
    """
    let X_t be a fold in the training set 
    and a_j be an alpha in alphas. Then 
    each thread takes the pair (X_t, a_j)
    for all j and t and computes the RMSE for 
    the path of all Lambda values in params.

    For a given alpha j and f folds,
    the RMSE for all Lambda values:

    [ [ SSE_1,j \in R^{1 x K} ]   -> (X_1, a_j)
       ...
      [ SSE_f,j \in R^{1 x K} ] ]  -> (X_f, a_j)

    Where  K is the number of Lambda values.

    If the average of the column j is taken,
    then it will effectively be the CV_error 
    for the pair (alpha_j, lambda_i) hyperparameters.
    """
    # X = X_train
    # y = y_train
    # num_folds = 5

    n = X_train.shape[0]
    fold_size = n // num_folds

    #shuffle the data
    all_index = list(range(n))
    rng.shuffle(all_index)

    X_train = X_train[all_index, :] # check to shuffle X
    y_train = y_train[all_index] # check to shuffle y!
    
    # tmp_seed = rng.randint(0, 2**31 - 1)
    if verbose:
        print("Seed for the ensemble: ", tmp_seed, ", fold size: ", fold_size)

    out = deque([])
    with mp.Pool( processes = ncores ) as pool:

        preout = deque([])
        for i in range(num_folds):

            test_idx = list(range(i * fold_size, (i + 1) * fold_size))
            train_idx = list(set(range(n)) - set(test_idx))

            X_train_t, X_test_t = X_train[train_idx, :], X_train[test_idx, :]
            y_train_t, y_test_t = y_train[train_idx], y_train[test_idx]

            for theta_j in full_grid:

                errors = pool.apply_async(
                    error_fn, 
                    (theta_j, base_model, i, 
                     X_train_t, X_test_t, 
                     y_train_t, y_test_t, 
                     verbose, tmp_seed) # seed fixed for the ensemble only
                )
                preout.append(errors)

        for errors in preout:
            out.append(errors.get())

    return list(out)

def get_best_params(all_errors, n = 1):
    # all_errors = out
    """
    Fold_alpha has the following structure:
    [  theta_{1,j}, ..., theta_{f,j}  ] 

    Fold error has the following structure for theta_j:
    [ [ SSE_{1,j} \in R^{1 x K} ]   -> theta_{1,j}
       ...
      [ SSE_{f,j} \in R^{1 x K} ] ]  -> theta_{f,j}
    
    For column k:
     (1/n) \Sum_{f = 1}^K  SSE_{f,j,k} = (1/n) \Sum_{i = 1}^n e^2_{i,j,k} =: MSE_{j,k}

    where f is the fold index, j is the hyperparameter index, and k 
    the lambda value index. This reduces to

    {.., theta_j: [ MSE_{j,1} ... MSE_{j,K} ], ...}

    We find the theta_j that has the lowest MSE. Alternatively it 
    we can also get the sqrt
    """
    # getting row-wise average of the errors
    dict_errs = {}

    # (1/n) \Sum_{f = 1}^K  SSE_{i,j,k}
    for (e_fj, theta_j) in all_errors:
        if isinstance(theta_j, list):
            # make it hashable if it is a list
            theta_j = tuple(theta_j)  

        if theta_j not in dict_errs:
            dict_errs[theta_j] = (e_fj/n)

        else:
            dict_errs[theta_j] += (e_fj/n)

    best_theta_j = None
    min_rmse = np.inf

    for theta_j, ave_e_j in dict_errs.items():
        tmp_cv_err = np.min(np.sqrt(ave_e_j))

        if tmp_cv_err < min_rmse:
            min_rmse = tmp_cv_err
            best_theta_j = theta_j
            # print(best_theta_j, best_lam, min_rmse)

    return best_theta_j, min_rmse

def ISLEPathCV(base_model, X_train, y_train, full_grid, folds, ncores,
               verbose = True, rng = None, tmp_seed = 1234):
    """
    Find the best set of hyperparameters for the ISLEPath
    using a cross-validation. The function returns
    the best hyperparameter set.

    Higltights: it parallelizes over all the folds and 
    hyperparameter values.


    Parameters
    ----------
    base_model : object
        The base model to be used for the ISLEPath.
    X_train : numpy.ndarray
        The training data.  
    y_train : numpy.ndarray
        The target values for the training data.    
    full_grid : list    
        The grid of hyperparameters to be used for the ISLEPath.
    folds : int
        The number of folds to be used for the cross-validation.    
    ncores : int
        The number of cores to be used for the parallelization.
    verbose : bool, optional
        If True, the function will print the progress of the cross-validation.
        The default is True.
    rng : numpy.random.RandomState, optional
        The random number generator to be used for randomly splitting the data.
        in cross-validation.
    tmp_seed : int, optional
        A temporary seed to be used for the ensemble generation. Every fold
        has the same ensemble seed and therefore the errors are more comparable.

    Returns
    -------
    tuple
        A tuple containing the best hyperparameters (nj, vj, lj, alpha) and
        the minimum average RMSE across all folds.
    """

    if len(full_grid) == 1:
        return full_grid[0]

    if verbose:
        print("Hyperparameter grid size: ", len(full_grid))
        print("Performing CV with ", folds, " folds")
        
    all_errors = get_parallel_errors(
        base_model, X_train, y_train, full_grid,
        rng, folds, ncores, 
        verbose, tmp_seed
    )

    (best_theta_j,
     min_rmse) = get_best_params(all_errors, n = X_train.shape[0])
    
    if verbose:
        print("CV best (eta, nu, leaves, alpha): ", best_theta_j)
        print("CV min average RMSE: ", min_rmse)

    return best_theta_j


def create_tree_param_grid(eta, nu, leaves, cv_sample = 100, rng = None):
    """
    return an array with combinations of hyperparameters
    for the ISLE ensemble. If cv_sample is lower than all possible
    combinations of hyperparameters, a random sample of size cv_sample
    is taken
    """
    param_size = len(eta) * len(nu) * len(leaves)
    
    if param_size <= cv_sample:
        grid = deque()
        for nj in eta:
            for vj in nu:
                for lj in leaves:
                    grid.append((nj, vj, lj))           
        grid = np.array(grid)
    else:
        grid = np.zeros((cv_sample, 3))
        grid[:,0] = rng.choice(eta, cv_sample)
        grid[:,1] = rng.choice(nu, cv_sample)
        # needs to be changed into integers
        grid[:,2] = rng.choice(leaves, cv_sample)

    return grid


def create_full_grid(isle, eta, nu, leaves, alphas, cv_sample, rng):

    assert all([a > 0 and a <= 1 for a in alphas]), "Alpha values must be between 0 and 1."

    if not isle:
        return [[None, None, None, a] for a in alphas]
    
    grid = create_tree_param_grid(eta, nu, leaves, cv_sample, rng)
    
    out_grid  = deque([])
    for alpha_j in alphas:
        for tree_parms_j in grid:
            nj, vj, lj = list(tree_parms_j)
            out_grid.append( [nj, vj, int(lj), alpha_j] )

    return out_grid



# high-dimensional sparse solutions

from collections import deque


import numpy as np
from qsin.utils import progressbar
from qsin.CD_dual_gap import (epoch_lasso_v2, update_beta_lasso_v2, dualpa,
                              epoch_enet_v2 , update_enet_v2, dualpa_elnet)


class Lasso:
    """
    lasso model.
    It assumes that the input data has been standardized. 
    If not, lasso does it nonetheless on (X,y)

    init beta is a zero vector
    """
    def __init__(self, 
                 max_iter=300, 
                 lam=0.1,
                 prev_lam = None,
                 fit_intercept  = True,
                 warm_start = False,
                #  beta = None,
                 tol = 0.001,
                 init_iter = 1,
                 copyX = False,):
        
        self.max_iter = max_iter
        self.lam = lam
        self.prev_lam = prev_lam
        self.tol = tol
        self.warm_start = warm_start

        self.fit_intercept = fit_intercept
        self.beta = np.array([])


        self.intercept = 0
        
        self.init_iter = init_iter
        self.copyX = copyX

        self.X = np.array([])
        self.ny2 = np.array([])
        self.r = np.array([])

        self.y = np.array([])

        self.x_bar = np.array([])
        self.y_bar = 0.0

        self.x_sd = np.array([])
        self.y_sd = 1.0

        self._verbose = False

        self.zero_thresh = 1e-15
        self.message_conv = ""

    def __str__(self):
        return f"Lasso(tol={self.tol}, lam={self.lam}, max_iter={self.max_iter})"
    
    def update_beta(self, c1, n, all_p):
        update_beta_lasso_v2(
            self.X, self.beta, self.lam, self.r,
            c1, n, all_p, self.zero_thresh
        )

    def initial_iterations(self, c1, all_p):

        n,_ = self.X.shape
        for _ in range(self.init_iter):
            # O(np)
            self.update_beta(c1, n, all_p)

    def sorted_init_active_set(self):
        """
        Get the active set A: non-zero coefficients

        Returns
        -------

        Aq: the active set as a deque sorted with respect to the indices\\
        A: the active set as a set, not necessarily sorted.
        """

        Aq = deque()
        A = set()
        # O(2p)
        for j, b in enumerate(self.beta):
            # if abs(b) >= self.zero_thresh:
            if b != 0.0:
                # O(2)
                A.add(j)
                Aq.append(j)

        return Aq, A

    def initial_active_set(self, c1, all_p):

        # few iterations of coordinate descent
        # O(np*T), where T is the number of initial iterations
        self.initial_iterations(c1, all_p)

        # we define an active set A as the set of indices
        # O(2p)
        return self.sorted_init_active_set()
    
    def set_Xy(self, X,y):
        """
        set X

        This is used under the 
        following logic

        if copy and x == 0:
            set_X
        elif not copy and x == 0:
            set_X (overwrite)
        elif copy and x != 0:
            do nothing
        elif not copy and x != 0:
            set_X (overwrite)
        """

        if self.copyX and len(self.X) != 0:
            pass

        else:
            # as contiguous array
            # makes the accessing of columns
            # faster. The same for residuals.
            # this fast block access is important
            # for multiple dot products in the
            # coordinate descent
            # XXX
            # asfortranarray makes a copy of the array
            self.X = np.asfortranarray(X) # O(np)
            
            # ascontiguousarray does not make a copy
            # but ensures that the array is contiguous
            self.y = np.ascontiguousarray(y.copy(), dtype=y.dtype) # O(n)
            self.center_Xy() # O(np)

            self.ny2 = np.dot(self.y, self.y) # O(n)
            self.r = self.y - self.X @ self.beta # O(np)
            self.r = np.ascontiguousarray(self.r, dtype=self.X.dtype) # O(n)
        
    def center_Xy(self):
        """
        center X and y
        """
        self.x_bar = np.mean(self.X, axis=0, dtype=self.X.dtype) # O(np)
        self.x_sd  = np.std(self.X , axis=0, dtype=self.X.dtype) # O(np)
        
        # avoid division by zero
        zero_sd_indices = self.x_sd == 0 # O(p)
        if np.any(zero_sd_indices): # O(p)
            self.x_sd[zero_sd_indices] = 1 # O(p)

        self.X = (self.X - self.x_bar) / self.x_sd # O(np)

        self.y_bar = np.mean(self.y) # O(n)
        self.y -= self.y_bar # O(n)

    def dual_gap(self, y):
        return dualpa(self.X, y, self.r, self.lam, self.beta, self.ny2)

    def get_sorted_complement(self, A, all_p):
        """
        get the sorted complement of the active set

        A is the active set and all_p is the set of all
        indices. It returns ordered list of indices
        """
        Ac_q = deque()
        # O(2p)  in average
        for j in all_p:
            if j not in A:
                Ac_q.append(j)

        return np.array(Ac_q) # O(p)
    
    def update_sorted_active_set(self, A, Ac_f, all_p):
        """
        update the active set based on the exclusion test
        it returns the new active set ordered with respect
        to the indices
        """
        Anew = deque()
        # O(3p) in average
        for j in all_p:
            if (j in A) or (j in Ac_f):
                Anew.append(j)

        return Anew
        
    def fit(self, X, y):
        # X = X_train
        # y = y_train
        # self.max_iter = 10000
        # self.set_params(max_iter = 100, lam = 0.1)
        
        n, p = X.shape
        c1 = 2 / n
        
        if not self.warm_start:            
            # zero initialization
            self.beta = np.zeros(p, dtype=X.dtype, order='F')

        # XXX, O(np)
        self.set_Xy(X, y)

        # few iterations of coordinate descent
        all_p = np.array(range(p))

        # O(2p + np*T_init) = O(np)
        Aq, A = self.initial_active_set(c1, all_p)

        
        if len(A) == 0:
            # if the active set is empty
            # then the model is converged
            if self.fit_intercept:
                self.set_intercept()
            return

        A_rr = np.array(Aq, dtype=np.int64)
        # print("Active set: ", A_rr)

        left_iter = self.max_iter - self.init_iter
        
        # O(T*c1*np + c2*p + c3*n) = O(np) for p >> n and T small,
        # where T is the number of left iterations 
        # and c1, c2, c3 are constants
        # Small T are practially possible with the warm starts
        # and thorough path of the lasso
        for i in range(left_iter):

            # O(n)
            xb_diff = np.zeros(n, dtype=np.float64)
            xb_new = np.zeros(n, dtype=np.float64)
            # O(np)
            self.cd_epoch(c1, n, A_rr, xb_new, xb_diff)

            # O(n)
            # checking convergence on the beta values only
            # might be not correct as in high dimensions (p > n) there 
            # does not exist a unique solution (i.e., X'X is singular). 
            # However, there is a unique solution for XB. so, we can check
            # the convergence on the XB values
            # ref: Sparsity, the Lasso, and Friends 
            # (section 3.2), Ryan Tibshirani, 2017.
            max_updt = np.max(np.abs(xb_diff))
            w_max = np.max(np.abs(xb_new))

            if self._verbose:
                print("iteration:", i,"Max update: ", max_updt, "Max weight: ", w_max)

            if w_max == 0 or max_updt/w_max < self.tol:
                
                # O(3p)
                Ac_arr = self.get_sorted_complement(A, all_p)

                #TODO: streamline updt. beta and exclusion test
                # O(np)
                self.update_beta(c1, n, all_p)

                # O(np + p) = O(np)
                # A_c that failed the exclusion test
                Ac_f_arr = self.exclusion_test(self.X, c1, Ac_arr)
                Ac_f = set(Ac_f_arr)
                

                if len(Ac_f_arr) == 0:
                    # it means that all
                    # coefficients follow the
                    # KKT conditions
                    if self._verbose:
                        print('kkt, finished at iteration: ', i)
                    break

                else:
                    # O(np + n + p) = O(np)
                    w_d_gap = self.dual_gap(self.y)
                    if w_d_gap < self.tol:
                        if self._verbose:
                            print('dual, finished at iteration: ', i)
                        break

                    else:
                        # O(3p)
                        Anew = self.update_sorted_active_set(A, Ac_f, all_p)
                        # O(2p)
                        A = set(Anew)
                        A_rr = np.array(Anew)

        if i == left_iter - 1:
            # if the iterations reach this point,
            # it means that there is still an active set.
            # then, the model did not converge
            self.message_conv = f"{self.__str__()} did not converge. Try to increase either max_iter or tol params."

        if self.fit_intercept:
            self.set_intercept()

    def set_intercept(self):
        # this scaling will consider the unscaled
        # data. The optimization was done assuming the
        # scaled data. 
        # If x_sd ever had a zero value, then
        # it was replaced by 1.
        self.beta /= self.x_sd
        self.intercept = self.y_bar - np.dot(self.x_bar, self.beta) # O(p)

        if abs(self.intercept) < self.zero_thresh:
            # if the intercept is close to zero, then
            # we can set it to zero
            self.intercept = 0.0

    def cd_epoch(self, c1, n, chosen_ps, s_new, s_diff):
        epoch_lasso_v2(
            self.X, self.beta, self.lam, self.r,
            c1, n, chosen_ps, s_new, s_diff, self.zero_thresh
        )

    def exclusion_test(self, X, c1, A_c_arr):
        """
        Exclusion test (SLS book, page 114)

        The optimization problem dictates that:

        X^T * r = lam * gamma
        
        gamma being partial of ||B||_1 and whose
        subgradient is given by:

        gamma_j = + 1     if beta_j > 0 (sign of beta_j)\\
        gamma_j = - 1     if beta_j < 0 (sign of beta_j)\\
        gamma_j = (-1, 1) if beta_j = 0 (sign of 0, undefined)\\

        Since the range of gamma_j is between -1 and 1,\\
        then the range of lam * gamma_j is between -lam and lam.\\
        So, the range of Xj^T * r is between -lam and lam.

        Leading the following set of inequalities:
        -lam <= Xj^T * r <= lam

        which from:\\
        lam >= Xj^T * r\\
        lam >= -Xj^T * r

        Implies: 
        lam >= |Xj^T * r|

        Now, we can define the exclusion test as:
        lam > |Xj^T * r|
        whose strict inequiality test for beta_j = 0
        as +1 and -1 are not in the range of the subgradient

        If we pass this test over the omited  variables
        that were supposed to be zero and fails, it means
        they were actually non-zero and should be included
        in the active set.
        """

        if len(A_c_arr) == 0:
            return A_c_arr
        
        # exclusion test (SLS book, page 114) based on
        # KTT conditions

        # O(np)
        e_test = c1 * np.abs( np.matmul( X.T[A_c_arr, :], self.r) ) >= self.lam
        # those that are in the new active set
        # are those who did not pass the KTT conditions.
        # true means that they failed the exclusion
        # test and should be included in the active set
        
        return A_c_arr[e_test] # O(p)

    def predict(self, X):
        if self.beta is None:
            raise Exception("Model has not been fitted yet.")
        
        return X @ self.beta + self.intercept
    
    def score(self, X, y):
        y_pred = self.predict(X)
        return np.sqrt( np.mean( (y - y_pred)**2 ) )
    
    def set_params(self, **params):
        if "max_iter" in params:
            self.max_iter = int(params["max_iter"])

        if "lam" in params:
            self.lam = params["lam"]

        if "intercept" in params:
            self.intercept = params["intercept"]

        if "tol" in params:
            self.tol = params["tol"]

        if "warm_start" in params:
            self.warm_start = params["warm_start"]

        if "beta" in params:
            self.beta = params["beta"]

        if "prev_lam" in params:
            self.prev_lam = params["prev_lam"]

    def get_params(self):
        return {'max_iter': self.max_iter, 
                'lam': self.lam, 
                'intercept': self.intercept, 
                'tol': self.tol}
    
    def set_beta(self, beta):
        self.beta = beta

# region: ElasticNet
class ElasticNet(Lasso):
    def __init__(self, 
                 max_iter=300, 
                 alpha = 0.5, 
                 lam=0.1, 
                 prev_lam=None, 
                 fit_intercept=True, 
                 warm_start=False,
                 tol=1e-4,
                 init_iter=1,
                 copyX=False):
        # passes arguments to the Lasso class
        super().__init__(max_iter, lam, prev_lam, fit_intercept, warm_start,  tol, init_iter, copyX)

        self.alpha = alpha
        self.lam = lam

        # when lam or alpha
        # change, these values are
        # updated at set_params
        self.lam_alpha = alpha * lam
        self.lam_1_alpha = (1 - alpha) * lam

    def __str__(self):
        return f"ElasticNet(tol={self.tol}, lam={self.lam}, alpha={self.alpha}, max_iter={self.max_iter})"

    def set_lam_alpha(self, lam, alpha):
        self.lam_alpha = alpha * lam
        self.lam_1_alpha = (1 - alpha) * lam

    def set_params(self, **params):
        if "max_iter" in params:
            self.max_iter = int(params["max_iter"])

        if "lam" in params:
            self.lam = params["lam"]

            if "alpha" in params:
                self.alpha = params["alpha"]
            
            self.set_lam_alpha(self.lam, self.alpha)
            # self.lam_alpha = self.alpha * self.lam
            # self.lam_1_alpha = (1 - self.alpha) * self.lam

        if 'alpha' in params:
            self.alpha = params['alpha']

            if "lam" in params:
                self.lam = params["lam"]

            self.set_lam_alpha(self.lam, self.alpha)
            # self.lam_alpha = self.alpha * self.lam
            # self.lam_1_alpha = (1 - self.alpha) * self.lam

        if "intercept" in params:
            self.intercept = params["intercept"]

        if "tol" in params:
            self.tol = params["tol"]

        if "warm_start" in params:
            self.warm_start = params["warm_start"]

        if "beta" in params:
            self.beta = params["beta"]

        if "prev_lam" in params:
            self.prev_lam = params["prev_lam"]

    def cd_epoch(self, c1, n, chosen_ps, xb_new, xb_diff):
        epoch_enet_v2(self.X, self.beta, self.lam_1_alpha, self.lam_alpha, self.r,
                      c1, n, chosen_ps, xb_new, xb_diff, self.zero_thresh)

    def update_beta(self, c1, n, chosen_ps):
        update_enet_v2(self.X, self.beta, self.lam_1_alpha, self.lam_alpha, self.r,
                       c1, n, chosen_ps, self.zero_thresh )

    def dual_gap(self, y):
        return dualpa_elnet(self.r, self.X, y, self.beta, self.lam_1_alpha, self.lam_alpha, self.ny2)
    
    def exclusion_test(self, X, c1, A_c_arr):

        if len(A_c_arr) == 0:
            return A_c_arr

        # exclusion test (SLS book, page 114) based on
        # KTT conditions
        if self.lam_1_alpha == 0:
            elastic_term = 0
        else:
            elastic_term = (1/c1) * self.lam_1_alpha * self.beta[A_c_arr]

        elastic_test = c1 * np.abs( np.matmul( X.T[A_c_arr, :], self.r ) - elastic_term ) >= self.lam_alpha
        # those that are in the new active set
        # are those who did not pass the KTT conditions
        return A_c_arr[elastic_test]
# endregion

# region: Cross-validation

def k_fold_cv(X, y, model, num_folds):
    n, p = X.shape
    fold_size = n // num_folds
    mse_sum = 0

    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]

        model.fit(X_train, y_train)
        # y_pred = lasso.predict(X_test)
        mse_sum += model.score( X_test, y_test )

    return mse_sum / num_folds

def k_fold_cv_random(X, y, 
                     model, 
                     params,
                     num_folds = 3, 
                     sample = 100,
                     verbose = False,
                     seed = 123,
                     warm_starts = False
                     ):
    
    # model = Lasso()
    # X = X_train
    # y = y_train
    
    np.random.seed(seed=seed)
    
    all_params = params.keys()
    tested_params = np.ones((sample, len(params)))

    for n,k in enumerate(all_params):
        tested_params[:,n] = np.random.choice(params[k], sample)
    
    if warm_starts:
        # check index where 'lam' is in 
        # all_params
        idx = list(all_params).index('lam')
        # sort the tested_params by using the idx in the decreasing
        # order
        tested_params = tested_params[tested_params[:,idx].argsort()[::-1]]
        # tested_params = tested_params[tested_params[:,idx].argsort()]
    
    all_errors = []
    for vec in tested_params:
        # vec = tested_params[1]
        tmp_params = dict(zip(all_params, vec))

        if warm_starts and len(model.beta):
            model.set_params(**tmp_params, 
                             warm_start=True, 
                             beta=model.beta)
        else:
            model.set_params(**tmp_params)


        tmp_err = k_fold_cv(X, y, model, num_folds)
        all_errors.append([tmp_params, tmp_err])

        if verbose:
            print('Error: %s, tested params: %s' % (tmp_err, vec))

    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]
    if verbose:
        print("CV score: ", best_[1])

    return best_[0]

def lasso_path(X_train, y_train, lams, model, print_progress = True,
               extra_str = ""):
    """
    compute the lasso path based on the training set
    and  with errors based on the test set

    it also returns the intercepts
    for each lambda value in the path
    so that the model can be used for prediction
    with the intercepts
    """

    _,p = X_train.shape
    K  = len(lams)

    path = np.ones((p, K))
    intercepts = np.zeros(K)

    model.set_params(lam=lams[0])
    model.fit(X_train, y_train)
    # all values at the beginning of the path are zeros.
    # convergence warning are due to floating point errors
    # and it should be ignored
    if model.message_conv:
        model.message_conv = ""

    path[:,0] = model.beta
    intercepts[0] = model.intercept

    if print_progress:
        index_set = progressbar(range(1, len(lams)), "Computing lasso path: ", 40)
        
    else:
        index_set = range(1, len(lams))
    
    for i in index_set:

        model.set_params(lam = lams[i],
                          warm_start = True, 
                          prev_lam = None)
        model.fit(X_train, y_train)

        if model.message_conv:
            print(f"Warning: {extra_str}+{model.message_conv}")
            model.message_conv = ""

        path[:,i] = model.beta
        intercepts[i] = model.intercept

    return path, intercepts

# endregion




import time
import json
from copy import deepcopy
from collections import deque

from qsin.sparse_solutions_hd import ElasticNet, lasso_path
from qsin.utils import progressbar, get_lambdas

import numpy as np
from sklearn.tree import DecisionTreeRegressor

def Sm(X, y, f_m, sample_size, replace = False, rng = None):
    # f_m = f_0

    n,p = X.shape
    test_idx  = rng.choice(range(n), size = sample_size, replace = replace)
    # print("test_idx: ", test_idx)

    return X[test_idx,:], y[test_idx], f_m[test_idx]

def make_isle_ensemble(X_train, y_train, model, eta, nu,
                      M, rng = None, verbose = True):
    

    n_train = X_train.shape[0]
    train_sample_size = np.clip( int(n_train*eta), 1, n_train )
    if verbose:
        print("Starting ISLE ensemble")
        print("Random sample size for trees: ", train_sample_size)

    # initialize memory function
    f_m = np.repeat(np.mean(y_train), n_train)

    F_train = np.zeros((n_train,M))
    estimators = deque()

    pb = progressbar(range(M), "Computing trees: ", 40) if verbose else range(M)
    # O(MnT^4log(n))
    for i in pb:
        model.set_params(random_state = rng)
        # random sample the data, including the memory function
        # O(nT^4)
        X_sm, y_sm, f_sm = Sm(X_train, y_train, f_m, train_sample_size, replace=False, rng=rng)
        # fit the model. O(p n log(n)), 
        # where p is considered number of features.
        # p <= T^4 -> O(p n log(n)) <= O(T^4 n log(n))
        model.fit(X_sm + f_sm.reshape(-1,1), y_sm )
        
        # update memory function
        f_m = f_m + nu*model.predict(X_train)
        
        F_train[:,i] = model.predict(X_train)  # O(n)
        # model largest object is the tree
        # and this has a size of O(d) = O(1) for
        # small d
        estimators.append(deepcopy(model)) # O(1)

    return F_train, list(estimators) # O(M)

def make_init_model(max_features = None, max_depth = 5, max_leaves = 6, param_file = None):

    if param_file is None:
        return DecisionTreeRegressor(max_features = max_features, 
                                     max_depth = max_depth, 
                                     max_leaf_nodes = max_leaves)
    
    else:
        # read json file
        # param_file = './tree_params.txt'
        with open(param_file) as f:
            params = json.load(f)

        return DecisionTreeRegressor(max_features = max_features,
                                     max_depth = max_depth,
                                     max_leaf_nodes = max_leaves
                                     **params)

def make_F_test(X_test, estimators):
    M = len(estimators)
    F_test = np.zeros((X_test.shape[0], M))
    
    # O(nM)
    for i,m in enumerate(estimators):
        F_test[:,i] = m.predict(X_test) # O(n)
    return F_test


def get_new_path(estimators, path):
    """
    this path is based on the feature importances that
    the selected estimators have. For each lambda, there 
    is an ensemble of estimators and the new_path contains
    the average feature importances of the ensemble

    The new path is a p x K matrix instead of M x K

    Parameters
    ----------
    estimators : list
        List of decision tree regressors.
    path : numpy.ndarray
        The path of coefficients with shape (M,k)
        where M is the number of estimators and k is the number of lambda values

    Returns
    -------
    list
        The new path, which is a list of lists of feature indices.
        Each set corresponds to a lambda value.
        The length of the list is equal to the number of lambda values.
        Each set contains the indices of the features selected by the ensemble
        of estimators for that lambda value.
    """

    estimators = np.array(estimators)
    new_path = deque()

    for j in range(path.shape[1]):
        # first iteration of the path
        # all the coefficients are 0: no selection
        if j == 0:
            # we add an empty set for backward compatibility
            # with the elastic net path
            new_path.append([])
            continue

        # j = 2
        coeffs = path[:,j]
        coeffs_logic = coeffs != 0.0

        tmp_ensemble = estimators[coeffs_logic]

        I_k = set()
        for m in tmp_ensemble:
            I_k_m = set(m.tree_.feature[m.tree_.feature != -2])
            I_k |= I_k_m

        I_k = list(I_k)
        new_path.append(I_k)

    return list(new_path)

class ISLEPath:
    """
    Path solution for the ISLE algorithm
    """
    def __init__(self,
                 # ISLE Ensemble params
                 eta = 0.5,
                 nu = 0.1,
                 max_leaves = 2,
                 M = 1000,
                 max_features = None,
                 max_depth = None,
                 param_file = None,
                 rng = None,
                 make_ensemble = True,
                 
                 # Elastic Net params
                 fit_intercept = True,
                 max_iter = 1000,
                 alpha = 0.5,
                 tol = 0.0001,
                 zero_thresh = 1e-15,  # threshold for zero coefficients
                 # Path params
                 epsilon  = 0.0001,
                 K = 100,
                 # other params
                 verbose = True
                 ):
        
        # ISLE ensemble params
        self.eta = eta
        self.nu = nu
        self.max_leaves = max_leaves
        self.M = M
        self.max_features = max_features
        self.max_depth = max_depth
        self.param_file = param_file
        self.rng = rng
        self.make_ensemble = make_ensemble

        # Elastic Net params
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.alpha = alpha
        self.zero_thresh = zero_thresh # internal parameter for the elastic net

        self.tol = tol

        # Path params
        self.epsilon = epsilon
        self.K = K

        # other params
        self.verbose = verbose

        # internal variables
        self.estimators = None
        self.path = None
        self.intercepts = None

        self.mu = None
        self.sigma = None
        self.lambdas = []  # will be set in fit method (pruning)

    def set_params(self, **params):
        """
        Set parameters for the ISLEPath instance.
        """
        for key, value in params.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(f"Parameter {key} is not recognized.")

    def generate_ensemble(self, X, y):
        base_tree = make_init_model(
            max_leaves=self.max_leaves,
            max_features=self.max_features,
            max_depth=self.max_depth,
            param_file=self.param_file
        )

        start = time.time()
        T, self.estimators = make_isle_ensemble(X, y, base_tree, 
                                                self.eta, self.nu, self.M, 
                                                rng=self.rng, 
                                                verbose=self.verbose)
        end_isle = time.time() - start

        if self.verbose:
            print("Isle ensemble done: ", end_isle, " seconds")

        # centering T for Elastic Net
        self.set_mu_sigma(T)
        T = (T - self.mu)/self.sigma

        return T

    def elnet_pruning(self, T, y):

        extra_str = f"Ensemble(eta={self.eta}, nu={self.nu}, max_leaves={self.max_leaves})"

        self.set_lambdas(T, y)
        elnet = ElasticNet(fit_intercept = self.fit_intercept,
                           max_iter = self.max_iter,
                           init_iter = 1,
                           copyX = True,
                           alpha = self.alpha,
                           tol = self.tol,)
        
        elnet.zero_thresh = self.zero_thresh

        (self.path, 
         self.intercepts) = lasso_path(T, y, self.lambdas, elnet,
                                       print_progress = self.verbose,
                                       extra_str = extra_str)
        # the effect of the intercept when (X,y) are centered
        # is negligible.
        # print("Intercepts: ", self.intercepts)
    
    def set_mu_sigma(self, T):
        self.mu = np.mean(T, axis=0)
        self.sigma = np.std(T, axis=0)
        
        sd_zero = self.sigma == 0
        if np.any(sd_zero):
            self.sigma[sd_zero] = 1

    def set_lambdas(self, X, y):
        self.lambdas = get_lambdas(self.alpha, X, y, self.K, self.epsilon, verbose=self.verbose)
        # max_lam = 1.7468125572517783
        # self.lambdas =  np.logspace(np.log10(max_lam*self.epsilon), np.log10(max_lam), self.K, endpoint=True)[::-1]

    def fit(self, X, y):

        if self.make_ensemble:
            T = self.generate_ensemble(X, y)
            # no need for centering in y.
        else:
            # if we are not making an ensemble,
            # then T is just X
            T = X

        # Lasso will re-center (X,y) and then 
        # optimize for \beta. The \beta will account for the
        # uncentered by getting \beta_0 and 
        # \hat \beta = \beta/\sigma, where \sigma is the
        # original variance. \sigma = ones if X is standarized.

        # notice that if T and y are already centered,
        # then the intercept (\beta_0) should be close to 0.
        self.elnet_pruning(T, y)

    def predict(self, X):
        n,_ = X.shape
        K = len(self.lambdas)

        if self.make_ensemble:
            assert self.estimators is not None, "Model fitting is necessary"
            T_test = make_F_test(X, self.estimators)
            T_test = (T_test - self.mu)/self.sigma

        else:
            # if we are not making an ensemble,
            # then F_test is just X
            T_test = X

        y_pred = np.zeros((n,K))
        for j in range(K):
            y_pred[:,j] = self.intercepts[j] + T_test @ self.path[:, j]
        
        return y_pred

    def score(self, X, y, metric = 'rmse'):
        
        y_pred = self.predict(X)
        # y_pred \in R^{n,K}, e.g.,

        # y_pred = np.ones((3,5)) + np.random.normal(size=(3,5))
        # y = np.array([1,2,3]).reshape(-1,1)

        if metric == 'rmse':
            return np.sqrt(np.mean((y_pred - y.reshape(-1,1))**2, axis=0))
        
        elif metric == 'sse':
            # sum of squared errors 
            return np.sum((y_pred - y.reshape(-1,1))**2, axis=0)
        
        else:
            raise ValueError(f"Unknown metric: {metric}")
    


#!/usr/bin/env python

import sys
import platform


import setuptools
from distutils.core import setup


dependencies = [
    'numpy',
    'numba',
    'scikit-learn',
]

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(name = "qsin",
      version = '0.11.15',
      maintainer = 'Ulises Rosas',
      packages = ['qsin'],
      package_dir = {'qsin': 'src'},
      package_data = {'qsin': ['data/*']} ,
      include_package_data=True,
      install_requires = dependencies,
      zip_safe = False,
      entry_points = {
        'console_scripts': [
            'path_subsampling.py   = qsin.path_subsampling:main'
            ]
      },
      scripts=[
          './scripts/infer_qlls.jl',
          './scripts/sim_nets.R',
          './scripts/infer_nets_batches.jl',
          # './src/path_subsampling.py'
      ],
      classifiers = [
          'Programming Language :: Python :: 3'
      ]
    )



#!/usr/bin/env python3

import time
import argparse

import numpy as np

from qsin.isle_path import get_new_path, ISLEPath
from qsin.ISLEPathCV import create_full_grid, ISLEPathCV
from qsin.row_selection import row_selection, write_rows
from qsin.utils import (standardize_Xy, write_test_errors, 
                        split_data)

def max_features_type(value):
    try:
        return float(value) if value not in ['sqrt', 'log2'] else value
    
    except ValueError:
        raise argparse.ArgumentTypeError(f"Invalid value for --max_features: {value}. Must be a float, 'sqrt', or 'log2'.")

def main():

    # region: parse arguments
    parser = argparse.ArgumentParser(description="""
    Generate batches from ElasticNet path.
    """, 
    add_help=False,
    formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    req_args = parser.add_argument_group("Required arguments")
    req_args.add_argument("Xy_file", help="Path to the Xy input data file.")
    req_args.add_argument("CT_file", help="Path to the CT file.")

    opt_args = parser.add_argument_group("Optional arguments")
    opt_args.add_argument("-h","--help", action="help", help="Show this help message and exit.")
    opt_args.add_argument("--verbose", action="store_true", help="Whether to print verbose output.")
    opt_args.add_argument("--isle", action="store_true", help="Whether to use path from decision tree-based ISLE (i.e., ensemble learning).")
    opt_args.add_argument("--p_test", type=float, default=0.35, metavar="", help="Proportion of observations to use for testing.")
    opt_args.add_argument("--seed", type=int, default=12038, metavar="", help="Random seed.")
    


    isle_args = parser.add_argument_group("ISLE parameters (if --isle is used)")
    isle_args.add_argument("--eta", type=float, nargs='+', default=[0.5], metavar="", help="Proportion of observations to use in each tree.")
    isle_args.add_argument("--nu", type=float, nargs='+', default=[0.1], metavar="", help="Learning rate.")
    isle_args.add_argument("--max_leaf_nodes", type=int, nargs='+' , default=[6], metavar="", help="Maximum number of leaf nodes in the decision tree.")
    isle_args.add_argument("--cv_sample", type=int, default=100, metavar="", help="Number of randomly choose ISLE hyperparameter combinations")
    isle_args.add_argument("--max_depth", type=int, default=5, metavar="", help="Maximum depth of the decision tree.")
    isle_args.add_argument("--M", type=int, default=500, metavar="", help="Number of trees in the ensemble.")

    # default according Hastie et al. 2009, pp. 592
    isle_args.add_argument("--max_features", type=max_features_type, default=0.3, metavar="", help="Maximum proportion of features it is considered to grow nodes in a regression tree. It can also be 'sqrt' or 'log2'.")
    isle_args.add_argument("--param_file", type=str, default=None, metavar="", help="""JSON file with parameters for the decision tree
                           different from max_depth and mx_p. The decision trees are made using
                           sklearn's DecisionTreeRegressor. Then a complete list of parameters can be found
                           at https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html.""")

    elnet_args = parser.add_argument_group("Elastic Net parameters")
    elnet_args.add_argument("--alpha", type=float, nargs='+', default=[0.999], metavar="", help="Alpha value controling l1 and l2 norm balance in ElasticNet.")
    elnet_args.add_argument("--e", type=float, default=1e-4,metavar="", help="Epsilon value, which is used to calculate the minimum lambda (i.e., min_lambda =  max_lambda * e).")
    elnet_args.add_argument("--K", type=int, default=100, metavar="",help="Number of lambda values to test between max_lambda and min_lambda. ")
    elnet_args.add_argument("--tol", type=float, default=0.00001,metavar="", help="Tolerance for convergence.")
    elnet_args.add_argument("--max_iter", type=int, default=1000,metavar="", help="Maximum number of iterations.")
    elnet_args.add_argument("--wpath", action="store_true",  help="Write ElasticNet path. Warning: This can be large.")
    elnet_args.add_argument("--nstdy", action="store_true", help="Not standardize y. Standarizing y helps to numerical stability. ")
    elnet_args.add_argument("--cv", action="store_true", help="Use cross-validation to select the best lambda and alpha value.")
    elnet_args.add_argument("--folds", type=int, default=5, metavar="", help="Number of folds for cross-validation when cv is True.")
    elnet_args.add_argument("--ncores", type=int, default=1, metavar="", help="Number of cores to use for cross-validation.")


    batch_args = parser.add_argument_group("Row selection parameters (After the model is trained)")
    batch_args.add_argument("--prefix", type=str, default='batches', metavar="", help="Prefix for output files.")
    batch_args.add_argument("--factor", type=float, default=-1, metavar="", help="Proportion of row selection. If factor is -1, the proportion of row selected is based on test error.")
    batch_args.add_argument("--inbetween", type=int, default=5, metavar="",help="Number of in-between sets from the solution of the beginning of the path to the final row selection.")
    batch_args.add_argument("--check_spps", action="store_true", help="Check the number of unique species in the selected rows.")
    batch_args.add_argument("--ignore_conv", action="store_true", help="Ignore convergence of test errors. If True, the function will ignore the convergence of the test errors and return the index of the minimum test error.")
    batch_args.add_argument("--tol_test", type=float, default=1e-4, metavar="", help="""
                            Tolerance for convergence of test errors for all k solutions in the ElNet path to pick the best lambda_k.
                            If the test errors are not converging, then lambda_k with the minimum test error is picked.""")

    args = parser.parse_args()
    # print(args)

    # assert args.factor >= -1 and args.factor <= 1, "Factor must be between 0 and 1."
    assert args.inbetween >= 0, "Inbetween must be greater or equal to 0."
    assert args.p_test > 0 and args.p_test < 1, "Proportion of test samples must be between 0 and 1."
    assert args.e > 0, "Epsilon must be greater than 0."
    assert args.K > 0, "K must be greater than 0."
    assert args.tol > 0, "Tolerance must be greater than 0."
    assert args.max_iter > 0, "Maximum number of iterations must be greater than 0."
    assert args.seed >= 0, "Seed must be greater or equal to 0."
    assert args.M > 0, "Number of trees in the ensemble must be greater than 0."
    assert args.max_depth > 0, "Maximum depth of the decision tree must be greater than 0."
    assert (isinstance(args.max_features, float) and args.max_features > 0 and args.max_features <= 1) or args.max_features in ['sqrt', 'log2'], "Maximum proportion of features must be between 0 and 1 or 'sqrt' or 'log2'."
    assert args.prefix != "", "Prefix must not be empty."
    # endregion: parse arguments

    # O(nT^4) for reading only
    data = np.loadtxt(args.Xy_file, delimiter=',', skiprows=1)
    X, y = data[:, :-1], data[:, -1]
    n,p = X.shape # p (\approx T^4) used for row selection only
    rng = np.random.RandomState(seed=args.seed)


    num_test = int(n*args.p_test)
    start = time.time()
    # O(M T^4 nlog(n)) for ISLE ensemble

    # O(nT^4) for centering
    X_train, X_test, y_train, y_test = split_data(X, y, num_test=num_test, seed=args.seed)

    # improve stability of the algorithm
    X_train, X_test, y_train, y_test = standardize_Xy(X_train, y_train, X_test, y_test, args.nstdy)

    base_model = ISLEPath(
            # elastic net parameters
            fit_intercept = True,
            max_iter = args.max_iter,
            tol = args.tol,
            zero_thresh = 1e-12,  # threshold for zero coefficients
            # path parameters
            epsilon = args.e,
            K = args.K,
            # tree parameters
            M = args.M,
            max_features = args.max_features,
            max_depth = args.max_depth,
            param_file = args.param_file,
            make_ensemble = args.isle,
            verbose = False,
        )
    
    full_grid = create_full_grid(isle=args.isle, eta=args.eta, nu=args.nu,
                                  leaves=args.max_leaf_nodes, alphas=args.alpha, 
                                  cv_sample=args.cv_sample, rng=rng)


    (nj, vj, lj, alpha) = ISLEPathCV(base_model, X_train, y_train, full_grid, 
                                     args.folds, args.ncores, 
                                     verbose = args.verbose, rng=rng, tmp_seed=args.seed)

    rng = np.random.RandomState(seed=args.seed) # reinitialize rng for the ensemble
    base_model.set_params(eta=nj, nu=vj, max_leaves=lj, alpha=alpha, 
                          verbose=args.verbose, rng=rng)

    # O(npk) = O(nT^4) for fixed k 
    # or O(nM) if isle is True
    base_model.fit(X_train, y_train)
    end_lasso = time.time() - start
    
    path = base_model.path
    lambdas = base_model.lambdas

    if args.verbose:
        print("ISLE path done: ", end_lasso, " seconds")


    if args.wpath:
        # the first column is the lambda values
        # which adds one extra column into the path
        lam_path = np.concatenate((lambdas.reshape(-1, 1), path.T), axis=1)
        np.savetxt(args.prefix + "_elnetPath.csv", 
                   lam_path, 
                   delimiter=',',
                   comments='')
    # O(n)
    # write the test errors as well.
    test_rmse = base_model.score(X_test, y_test)
    if args.verbose:
        print("Min test RMSE: ", np.min(test_rmse), "at lambda: ",
              lambdas[np.argmin(test_rmse)])

    # print(base_model.intercepts)
    # sel = base_model.path[:,np.argmin(test_rmse)]
    # print(sel)
    # print(np.sum(sel != 0), "non-zero coefficients at the best lambda_k solution.")

    # test errors contain the first column with lambda values
    # and the second column with RMSE values.
    # this is compatible with the row selection. 
    # TODO: streamline from test_rmse
    test_errors = write_test_errors(args.prefix, test_rmse, lambdas)

    if args.isle:
        picked_file = args.prefix + "_overlappedBatches_isle.txt"
        
        # transform the path
        # O(kp) = O(kT^4) = O(T^4) for fixed k
        # it returns a list of sets
        path = get_new_path(base_model.estimators, path) 
        
    else:
        picked_file = args.prefix + "_overlappedBatches.txt"

    # overlapping batches
    # O(\rho T^4) for isle. If check_spps is True, it is O(T^4)
    rows_selected, j_opt = row_selection(path, test_errors, p,
                                  args.factor, args.inbetween, args.check_spps,
                                  args.CT_file, args.verbose, args.tol_test, 
                                  args.ignore_conv)
    
    if args.verbose:
        print(f" at lambda: {lambdas[j_opt]}.")
    # O(\rho T^4) where \rho is the fraction of non-zero coefficients
    write_rows(picked_file, rows_selected)


    if args.verbose:
        # print(test_errors)
        min_err_sel = len(rows_selected[-1]) # the best is the last one
        print("Number of rows selected at lambda_k solution: ", min_err_sel)

if __name__ == "__main__":
    main()




from sparse_solutions_hd import ElasticNet, Lasso
from qsin.utils import split_data

import numpy as np
import time
seed = 12038
np.random.seed(seed)
n = 600
p = 100000
lam=0.0001
alpha = 0.5
tol = 1e-5
X = np.random.randn(n, p)
y = np.random.randn(n)
n = len(y)

X_train, X_test, y_train, y_test = split_data(X, y, 200)
X_train = (X_train - np.mean(X_train, axis=0)) / np.std(X_train, axis=0)
X_test = (X_test - np.mean(X_test, axis=0)) / np.std(X_test, axis=0)


# self = Lasso(max_iter=1000, lam=lam, seed=seed, tol=tol, fit_intercept=True)
self = ElasticNet(max_iter=1000, lam=lam, seed=seed, tol=tol, alpha=alpha, fit_intercept=True)
self._verbose = False
start = time.time()
self.fit(X_train, y_train)
print("Time: ", time.time() - start)
print(np.sum(self.beta != 0))
test_error = self.score(X_test, y_test)
print("Test error: ", test_error)

from sklearn.linear_model import ElasticNet as skElasticNet
from sklearn.linear_model import Lasso as skLasso

sklasso = skElasticNet(alpha=lam/2, l1_ratio=alpha, max_iter=1000, tol=tol, fit_intercept=True)
# sklasso = skLasso(alpha=lam/2, max_iter=10000, tol=tol, fit_intercept=True)
start = time.time()
sklasso.fit(X_train, y_train)
print("\nSklearn Time: ", time.time() - start)
sktest_error = np.sqrt(np.mean((y_test - sklasso.predict(X_test))**2))
print("Sklearn Test error: ", sktest_error)
print(sklasso.coef_[sklasso.coef_ != 0].shape)
